/**************************************************************
*                                                            *
*    CUDA_Sobel.vcproj                                       *
* date       : 21.2.2006                                     *
* modified   : 1. 3.2006                                     *
* author     : Xie Yongming                                  *
* e-mail     : ymxie@cse.cuhk.edu.hk                         *
*                                                            *
**************************************************************/



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/gl.h>
#include <GL/glut.h>
#include "cutil.h"
#include "cuda_gl_interop.h"


////////////////////////////////////////////////////////////////////////////////
// kernels
#ifndef _SOBEL_KERNEL_H_
#define _SOBEL_KERNEL_H_

///////////////////////////////////////////////////////////////////////////////
__global__ void
Sobel_kernel( float* Sdata, float* data,unsigned int width, unsigned int height)
{
    unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
    unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;
    unsigned int i = (y*width+x)*3;
    if(x>0&&y>0&&x<(width-1)&&y<(height-1))
    {
	int i0 = ((y-1)*width+(x-1))*3;
	int i1 = ((y-1)*width+(x+1))*3;
	int i2 = (y*width+(x-1))*3;
        
	int i3 = (y*width+(x+1))*3;
	int i4 = ((y+1)*width+(x-1))*3;
	int i5 = ((y+1)*width+(x+1))*3;
        
        float as0 = Sdata[i1+0] -Sdata[i0+0] -2.0*Sdata[i2+0] +2.0*Sdata[i3+0] -Sdata[i4+0] +Sdata[i5+0] ;
        float as1 = Sdata[i1+1] -Sdata[i0+1] -2.0*Sdata[i2+1] +2.0*Sdata[i3+1] -Sdata[i4+1] +Sdata[i5+1] ;
        float as2 = Sdata[i1+2] -Sdata[i0+2] -2.0*Sdata[i2+2] +2.0*Sdata[i3+2] -Sdata[i4+2] +Sdata[i5+2] ;
        
	int ii0 = ((y-1)*width+(x-1))*3;
	int ii1 = ((y-1)*width+(x))*3;
	int ii2 = ((y-1)*width+(x+1))*3;
        
	int ii3 = ((y+1)*width+(x-1))*3;
	int ii4 = ((y+1)*width+(x))*3;
	int ii5 = ((y+1)*width+(x+1))*3;
        
	float ass0 = Sdata[ii0+0] +2.0*Sdata[ii1+0] +Sdata[ii2+0] -Sdata[ii3+0] -2.0*Sdata[ii4+0] -Sdata[ii5+0] ;
	float ass1 = Sdata[ii0+1] +2.0*Sdata[ii1+1] +Sdata[ii2+1] -Sdata[ii3+1] -2.0*Sdata[ii4+1] -Sdata[ii5+1] ;
	float ass2 = Sdata[ii0+2] +2.0*Sdata[ii1+2] +Sdata[ii2+2] -Sdata[ii3+2] -2.0*Sdata[ii4+2] -Sdata[ii5+2] ;
        
        data[i]   = abs(as0)+abs(ass0);
        data[i+1] = abs(as1)+abs(ass1);
        data[i+2] = abs(as2)+abs(ass2);
    }
}

#endif // #ifndef _SOBEL_KERNEL_H_
///////////////////////////////////////////////////////////////////////////////
//
// Opengl
void initGL();
void C_VO2( GLuint* vbo);
void C_VO2_Image(GLuint* vbo,float* Image_data);
void U_VO_Image(GLuint vbo,float *Image_data);
void D_VO( GLuint* vbo);
void Show();

// OpenCV
extern int getFrame(float *img);


// Cuda functionality
void  Cuda_Sobel( GLuint vbo,GLuint vbo1);

////////////////////////////////////////////////////////////////////////////////
int window_width = 640;
int window_height = 240;
int image_width = 320;
int image_height = 240;
int image_depth = 3;

GLuint vbo1[2];
GLuint vbo_data1[2];

float  *B1;
bool edge = false;
bool play = false;

int vindex = 0;
int vnextIndex;

////////////////////////////////////////////////////////////////////////////////
void idlefunc(void) {
	glutPostRedisplay();
}

void key(unsigned char k, int x, int y)
{
  
    if(k==27) exit(0);
    if(k==' ')
    {
        play = !play;
        printf("Play: %d\n", play);
    }
    if(k=='e')
    {
        edge = !edge;
        printf("Edge: %d\n", edge);      
    }

    glutPostRedisplay();
}

void keyInfo()
{
    printf("\nKeyboard Comands:\n");
    printf("<space>: starts/stops getting video frames\n");
    printf("e      : starts/stops edge detection \n");
    printf("<esc>  : quits program\n");
    fflush(stdout);
}


////////////////////////////////////////////////////////////////////////////////
// main
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv) {
         
 	// Create GL 
	glutInit( &argc, argv);
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE);
	glutInitWindowSize( window_width, window_height);
	glutCreateWindow( "CUDA Image Processing");

	// initialize GL
	initGL();

	// register callbacks
	glutDisplayFunc( Show);
	glutKeyboardFunc( key);
	glutIdleFunc(idlefunc);

        getFrame(B1);

	// create VBO
	C_VO2( vbo1);
        C_VO2_Image(vbo_data1,B1);

        keyInfo();
	// start rendering mainloop
	glutMainLoop();

	D_VO( vbo1);
	D_VO( vbo_data1);
    CUT_EXIT(argc, argv);
}



////////////////////////////////////////////////////////////////////////////////
//Run Cuda Sobel
////////////////////////////////////////////////////////////////////////////////
void Cuda_Sobel( GLuint vbo,GLuint vbo1)
{ 
    // execute the kernel
    dim3 block(8, 8, 1);
    dim3 grid(image_width / block.x, image_height / block.y, 1);
    float *dptr;
    CUDA_SAFE_CALL(cudaGLMapBufferObject( (void**)&dptr, vbo));
    float *dptr1;
    CUDA_SAFE_CALL(cudaGLMapBufferObject( (void**)&dptr1, vbo1));

    Sobel_kernel<<< grid, block>>>(dptr1, dptr, image_width, image_height);
    
   // unmap buffer object
    CUDA_SAFE_CALL(cudaGLUnmapBufferObject( vbo));
    CUDA_SAFE_CALL(cudaGLUnmapBufferObject( vbo1));

}


void  initGL()
{
	
	B1 = (float *)malloc(image_width*image_height*3*sizeof(float));

	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
	/* Problem: glewInit failed, something is seriously wrong. */
		printf("Error: %s\n", glewGetErrorString(err));
	}

    // viewport
    glViewport( 0, 0, window_width, window_height);

     CUT_CHECK_ERROR_GL();

}


void C_VO2(GLuint* vbo)
{
    // create buffer object
    glGenBuffers( 2, vbo);
    glBindBuffer( GL_ARRAY_BUFFER, vbo[0]);

    // initialize buffer object
    unsigned int size = image_width * image_height * image_depth * sizeof( float);
    float *data = (float *) malloc(size);
    glBufferData( GL_ARRAY_BUFFER, size, data, GL_DYNAMIC_DRAW);
    
    glBindBuffer( GL_ARRAY_BUFFER, vbo[1]);
    glBufferData( GL_ARRAY_BUFFER, size, data, GL_DYNAMIC_DRAW);
    free(data);

    glBindBuffer( GL_ARRAY_BUFFER, 0);

    // register buffer object with CUDA
    CUDA_SAFE_CALL(cudaGLRegisterBufferObject(vbo[0]));
    CUDA_SAFE_CALL(cudaGLRegisterBufferObject(vbo[1]));

    CUT_CHECK_ERROR_GL();
}

void C_VO2_Image(GLuint* vbo,float *Image_data)
{
    // create buffer object
    glGenBuffers( 2, vbo);
    glBindBuffer( GL_ARRAY_BUFFER, vbo[0]);

    // initialize buffer object
    unsigned int size = image_width * image_height * image_depth * sizeof( float);
    float *data = (float *) malloc(size);

	for(int i =0; i<image_width * image_height * image_depth ;i++)
            data[i] = Image_data[i];

    glBufferData( GL_ARRAY_BUFFER, size, data, GL_DYNAMIC_DRAW);
    
    glBindBuffer( GL_ARRAY_BUFFER, vbo[1]);
    glBufferData( GL_ARRAY_BUFFER, size, data, GL_DYNAMIC_DRAW);
    free(data);
  
    glBindBuffer( GL_ARRAY_BUFFER, 0);
    // register buffer object with CUDA
    CUDA_SAFE_CALL(cudaGLRegisterBufferObject(vbo[0]));
    CUDA_SAFE_CALL(cudaGLRegisterBufferObject(vbo[1]));

    CUT_CHECK_ERROR_GL();
}


void U_VO_Image(GLuint vbo,float *Image_data)
{
    glBindBuffer( GL_ARRAY_BUFFER, vbo);

    // buffer object
    float* data = (float*)glMapBufferARB(GL_ARRAY_BUFFER, GL_WRITE_ONLY);	
    if (data)
    {
	for(int i =0; i<image_width * image_height * image_depth ;i++)
            data[i] = Image_data[i];
        glUnmapBufferARB(GL_ARRAY_BUFFER) ;
    }
    glBindBuffer( GL_ARRAY_BUFFER, 0);

    CUT_CHECK_ERROR_GL();
}

void D_VO( GLuint* vbo)
{
    glDeleteBuffers( 2, vbo);
    CUDA_SAFE_CALL(cudaGLRegisterBufferObject(vbo[0]));
    CUDA_SAFE_CALL(cudaGLRegisterBufferObject(vbo[1]));
}

void Show()
{
    vindex = (vindex + 1) % 2;
    vnextIndex = (vindex + 1) % 2;
    if (play)
    {
        getFrame(B1);
        U_VO_Image(vbo_data1[vnextIndex], B1);
    }
    if (edge)
        Cuda_Sobel(vbo1[vindex], vbo_data1[vindex]);	// run CUDA Sobel kernel
        
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // set view matrix
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // render from the vbo
    glBindBuffer(GL_ARRAY_BUFFER, vbo1[vindex]);
    float* P0 = (float*)glMapBufferARB(GL_ARRAY_BUFFER, GL_READ_ONLY);	
    glRasterPos2i(-1,-1);
    glDrawPixels(image_width,image_height,GL_RGB,GL_FLOAT,P0);
    glUnmapBufferARB(GL_ARRAY_BUFFER) ;

    glBindBuffer(GL_ARRAY_BUFFER, vbo_data1[vindex]);
    float* P1 = (float*)glMapBufferARB(GL_ARRAY_BUFFER, GL_READ_ONLY);	
    glRasterPos2i(0,-1);
    glDrawPixels(image_width,image_height,GL_RGB,GL_FLOAT,P1);
    glUnmapBufferARB(GL_ARRAY_BUFFER) ;

    glutSwapBuffers();
}



